-------------------------------------------------------------------------------
--
-- VGA video generator
--
-- Uses the vga_timing module to generate hsync etc.
-- Massages the RAM address and requests cycles from the memory controller
-- to generate video using one byte per pixel
--
-- Cristian Soviani, Dennis Lim, and Stephen A. Edwards
--
-- Modified for W4840 Lab #5 by: Eric Li, Ke Xu, Winston Chao
-------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

entity vga is
  port (
    clk            : in std_logic;
    pix_clk        : in std_logic;
    rst            : in std_logic;
    video_data     : in std_logic_vector(15 downto 0);
    video_addr     : out std_logic_vector(19 downto 0);
    video_req      : out std_logic;
    VIDOUT_CLK     : out std_logic;
    VIDOUT_RCR     : out std_logic_vector(9 downto 0);
    VIDOUT_GY      : out std_logic_vector(9 downto 0);
    VIDOUT_BCB     : out std_logic_vector(9 downto 0);
    VIDOUT_BLANK_N : out std_logic;
    VIDOUT_HSYNC_N : out std_logic;
    VIDOUT_VSYNC_N : out std_logic);
end vga;

architecture Behavioral of vga is

  constant H_ACTIVE      : integer := 640;
  constant V_ACTIVE      : integer := 480;
  constant H_TOTAL       : integer := 800;
  constant V_TOTAL       : integer := 524;

  -- Fast low-voltage TTL-level I/O pad with 12 mA drive
  
  component OBUF_F_12
    port (
      O : out STD_ULOGIC;
      I : in STD_ULOGIC);
  end component;

  -- Basic edge-sensitive flip-flop
  
  component FD
    port (
      C : in std_logic;
      D : in std_logic;
      Q : out std_logic);
  end component;

  -- Force instances of FD into pads for speed
  
  attribute iob : string;
  attribute iob of FD : component is "true";

  component vga_timing
    port (
      h_sync_delay         : out std_logic;
      v_sync_delay         : out std_logic;
      blank                : out std_logic;
      vga_ram_read_address : out std_logic_vector (19 downto 0);
      pixel_clock          : in std_logic;
      reset                : in std_logic;
      char_line            : out std_logic_vector(2 downto 0);
      pix_ld               : out std_logic;
      vreq                 : out std_logic;
      char_ld              : out std_logic;
      char_sh              : out std_logic;
      px_count             : out std_logic_vector(10 downto 0);
      ln_count             : out std_logic_vector(9 downto 0)
    );
  end component;

  -- Add ascii rom component
  component asciirom
    port (
      clk     : in std_logic;
      addr    : in std_logic_vector(9 downto 0);
      data    : out std_logic_vector(7 downto 0)
    );
  end component;

  -- Add sprite rom component
  component spriterom
    port (
      clk     : in std_logic;
      addr    : in std_logic_vector(6 downto 0);
      data    : out std_logic_vector(15 downto 0)
    );
  end component;

  signal r                      : std_logic_vector (9 downto 0);
  signal g                      : std_logic_vector (9 downto 0);
  signal b                      : std_logic_vector (9 downto 0);
  signal blank                  : std_logic;
  signal hsync                  : std_logic;
  signal vsync                  : std_logic;
  signal vga_ram_read_address   : std_logic_vector(19 downto 0);
  signal vreq                   : std_logic;
  signal vga_shreg              : std_logic_vector(15 downto 0);

  signal ascii_addr             : std_logic_vector(9 downto 0);
  signal char_data              : std_logic_vector(7 downto 0);
  signal color_value            : std_logic_vector(9 downto 0);
  signal char_line              : std_logic_vector(2 downto 0);
  signal char_shreg             : std_logic_vector(7 downto 0);
  signal pix_ld                 : std_logic;
  signal char_ld                : std_logic;
  signal char_sh                : std_logic;
  signal color_val              : std_logic_vector(2 downto 0);
  signal color_val_next         : std_logic_vector(2 downto 0);
  signal color_val_next_next    : std_logic_vector(2 downto 0);

  signal sprite_addr            : std_logic_vector(6 downto 0);
  signal sprite_data            : std_logic_vector(15 downto 0);

  signal sprite_px_toggle       : std_logic;
  signal sprite_px_counter      : std_logic_vector(3 downto 0);
  signal sprite_ln_toggle       : std_logic;
  signal sprite_ln_counter      : std_logic_vector(3 downto 0);
  signal px_count               : std_logic_vector(10 downto 0);
  signal ln_count               : std_logic_vector(9 downto 0);
  signal sprite_reg             : std_logic_vector(15 downto 0);

  signal sprite_x               : std_logic_vector(10 downto 0);
  signal sprite_y               : std_logic_vector(9 downto 0);
  signal sprite_encoding        : std_logic_vector(2 downto 0);
  signal sprite_color           : std_logic_vector(8 downto 0);

  signal r_tile_val             : std_logic_vector(9 downto 0);
  signal g_tile_val             : std_logic_vector(9 downto 0);
  signal b_tile_val             : std_logic_vector(9 downto 0);
  signal r_spr_val              : std_logic_vector(9 downto 0);
  signal g_spr_val              : std_logic_vector(9 downto 0);
  signal b_spr_val              : std_logic_vector(9 downto 0);


begin

  -- vga timing instance
  st : vga_timing port map (
    pixel_clock => pix_clk,
    reset => rst,
    h_sync_delay => hsync,
    v_sync_delay => vsync,
    blank => blank,
    vga_ram_read_address => vga_ram_read_address,
    char_line => char_line,
    pix_ld => pix_ld,
    vreq => vreq,
    char_ld => char_ld,
    char_sh => char_sh,
    px_count => px_count,
    ln_count => ln_count
  );

  -- FIXME: This should be disabled during blanking to reduce memory traffic

  -- Generate video_req (to the RAM controller) by delaying vreq by
  -- a cycle synchronized with the pixel clock
  
  process (clk)
  begin
    if clk'event and clk='1' then
      video_req <= pix_clk and vreq;
    end if;
  end process;

  -- The video address is the upper 19 bits from the VGA timing generator
  -- because we are using two pixels per word and the RAM address counts words

  video_addr <= '0' & vga_ram_read_address(19 downto 1);

  -- The video shift register: either load it from RAM or shift it up a byte
  -- The load and shift signals come from vga_timing
  process (pix_clk)
  begin
    if pix_clk'event and pix_clk='1' then
      if char_ld = '1' then
        vga_shreg <= video_data;
      elsif char_sh = '1' then
        -- Shift the low byte of read video data into the high byte
        vga_shreg <= vga_shreg(7 downto 0) & "00000000";
      end if;
    end if;
  end process;

  process (pix_clk)
  begin
    if ( pix_clk'event and pix_clk = '1' ) then
      if ( px_count = 500 and ln_count = 0 ) then
        sprite_x <= video_data(10 downto 0);
      end if;
      if ( px_count = 600 and ln_count = 0 ) then
        sprite_y <= video_data(9 downto 0);
      end if;
      if ( px_count = 700 and ln_count = 0 ) then
        sprite_encoding <= video_data(2 downto 0);
      end if;
    end if;
  end process;



  -- Generate the address line into the ROM
  ascii_addr <= vga_shreg(14 downto 8) & char_line;

  -- Use the top 3 bits of the vga_shreg as color bits
  color_val_next_next <= vga_shreg(14 downto 12);

  -- Instance of the ascii ROM
  char_loader : asciirom port map (
    clk => pix_clk,
    addr => ascii_addr,
    data => char_data
  );


  -- Instance of the sprite ROM
  sprite_loader : spriterom port map (
    clk => pix_clk,
    addr => sprite_addr,
    data => sprite_data
  );

--  sprite_x <= "00000000000";

--  process ( pix_clk )
--  begin
--    if ( pix_clk'event and pix_clk = '1' ) then
--      if ( px_count = "00000000000" and ln_count = "0000000000" ) then
--        if ( sprite_x = H_ACTIVE ) then
--          sprite_x <= "00000000000";
--          sprite_y <= sprite_y + 16;
--        else
--          sprite_x <= sprite_x + 2;
--        end if;
--        if ( sprite_y = V_ACTIVE ) then
--          sprite_y <= "0000000000";
--        end if;
--      end if;
--    end if;
--  end process;

  sprite_color <= "111101000";
  --sprite_encoding <= px_count(3 downto 2) + px_count(2);

  -- Generate sprite_addr 
  sprite_addr <= sprite_encoding & sprite_ln_counter;


  -- Sprite pixel toggle
  process (pix_clk)
  begin
    if ( rst = '1' ) then
      sprite_px_toggle <= '0';
    else 
      if ( pix_clk'event and pix_clk = '1' ) then
        if ( (sprite_x = px_count + 1) or 
             (sprite_x = "00000000000" and px_count = H_TOTAL - 1) ) then
          sprite_px_toggle <= '1';
        elsif ( sprite_px_counter = "0000" ) then
          sprite_px_toggle <= '0';
        end if;
      end if;
    end if;
  end process;

  -- Sprite pixel counter
  process (pix_clk)
  begin
    if ( rst = '1' ) then
      sprite_px_counter <= "0000";
    else 
      if ( pix_clk'event and pix_clk = '1' ) then
        if ( sprite_px_toggle = '1' ) then
          sprite_px_counter <= sprite_px_counter + 1;
        end if;
      end if;
    end if;
  end process;

  -- Sprite line toggle
  process (pix_clk)
  begin
    if ( rst = '1' ) then
      sprite_ln_toggle <= '0';
    else
      if ( pix_clk'event and pix_clk = '1' ) then
        if ( (sprite_y = ln_count + 1 and px_count = H_TOTAL - 1) or 
             (sprite_y = "0000000000" and ln_count = V_TOTAL - 1 and px_count = H_TOTAL - 1 ) ) then
          sprite_ln_toggle <= '1';
        elsif ( sprite_ln_counter = "0000" and px_count = H_TOTAL - 1 ) then
          sprite_ln_toggle <= '0';
        end if;
      end if;
    end if;
  end process;

  -- Sprite line counter
  process (pix_clk)
  begin
    if ( rst = '1' ) then
      sprite_ln_counter <= "0000";
    else 
      if ( pix_clk'event and pix_clk = '1' ) then
        if ( sprite_ln_toggle = '1' and ((sprite_y = 0 and px_count = H_TOTAL - 1) or (not(sprite_y = 0) and px_count = H_TOTAL - 2)) ) then
          sprite_ln_counter <= sprite_ln_counter + 1;
        end if;
      end if;
    end if;
  end process;

  -- delay the color value by 2 ticks
  process (pix_clk)
  begin
    if ( pix_clk'event and pix_clk = '1' ) then
      color_val <= color_val_next; 
      color_val_next <= color_val_next_next;
    end if;
  end process;

  -- 16 bit shift register to hold each line of pixels from a sprite
  process (pix_clk)
  begin
    if ( pix_clk'event and pix_clk = '1' ) then
      if ( (px_count = sprite_x - 1) or
           (sprite_x = 0 and px_count = H_TOTAL - 1 and sprite_ln_toggle = '1') ) then
        sprite_reg <= sprite_data;
      elsif ( sprite_px_toggle = '1' ) then
        sprite_reg <= sprite_reg(14 downto 0) & '0';
      end if;
    end if;
  end process;

  -- 8 bit shift register to hold each line of pixels from a char
  process (pix_clk)
  begin
    if ( pix_clk'event and pix_clk = '1' ) then
      if ( pix_ld = '1' ) then
        char_shreg <= char_data;
      else
        char_shreg <= char_shreg(6 downto 0) & '0';
      end if;
    end if;
  end process;

  -- use the MSB of the 8 bit sr to plot the pixel
--  color_value <= "1110000000" when char_shreg(7) = '1' else
--                 "0000000000";
  r_tile_val <= color_val(2) & color_val(2) & color_val(2) & "1111111" when char_shreg(7) = '1' else "0000000000";
  g_tile_val <= color_val(1) & color_val(1) & color_val(1) & "1111111" when char_shreg(7) = '1' else "0000000000";
  b_tile_val <= color_val(0) & color_val(0) & color_val(0) & "1111111" when char_shreg(7) = '1' else "0000000000";

  r_spr_val <= r_tile_val when sprite_reg(15) = '0' else sprite_color(8 downto 6) & "1111111";
  g_spr_val <= g_tile_val when sprite_reg(15) = '0' else sprite_color(5 downto 3) & "1111111";
  b_spr_val <= b_tile_val when sprite_reg(15) = '0' else sprite_color(2 downto 0) & "1111111";
  
  --r_spr_val <= sprite_color(8 downto 6) & "1111111";
  --g_spr_val <= sprite_color(5 downto 3) & "1111111";
  --b_spr_val <= sprite_color(2 downto 0) & "1111111";
  
  r <= r_spr_val when ( sprite_px_toggle = '1' and sprite_ln_toggle = '1' ) else r_tile_val;
  g <= g_spr_val when ( sprite_px_toggle = '1' and sprite_ln_toggle = '1' ) else g_tile_val;
  b <= b_spr_val when ( sprite_px_toggle = '1' and sprite_ln_toggle = '1' ) else b_tile_val;

  -- Video clock I/O pad to the DAC
  
  vidclk : OBUF_F_12 port map (
    O => VIDOUT_clk,
    I => pix_clk);

  -- Control signals: hsync, vsync, and blank

  hsync_ff : FD port map (
    C => pix_clk,
    D => not hsync,
    Q => VIDOUT_HSYNC_N );

  vsync_ff : FD port map (
    C => pix_clk,
    D => not vsync,
    Q => VIDOUT_VSYNC_N );

  blank_ff :  FD port map (
    C => pix_clk,
    D => not blank,
    Q => VIDOUT_BLANK_N );

  -- Three digital color signals
  
  rgb_ff : for i in 0 to 9 generate
    
    r_ff : FD port map (
      C => pix_clk,
      D => r(i),
      Q => VIDOUT_RCR(i) );
    
    g_ff : FD port map (
      C => pix_clk,
      D => g(i),
      Q => VIDOUT_GY(i) );
    
    b_ff : FD port map (
      C => pix_clk,
      D => b(i),
      Q => VIDOUT_BCB(i) );
    
  end generate;
  
end Behavioral;
